import os as _os
import tarfile as _tarfile
import zipfile as _zipfile
import urllib as _urllib
import tensorflow as _tf


def download_file(source='https://www.download.com/file.zip', filepath='./data/file.zip') -> None:
    """
    Download a file from source location and store it at the specified destination location.

    Args:
        source: the source URL
        filepath: the destination path and filename
    """
    path = _os.path.dirname(filepath)
    if path:
        _os.makedirs(path, exist_ok=True)
    else:
        _os.path.join('./', filepath)

    # noinspection PyUnresolvedReferences
    _urllib.request.urlretrieve(source, filepath)
    print(f'Download of {source} completed.')


def extract_tgz(filepath='./data/file.tar.gz', folder='./data') -> None:
    """
    Extracts a .tgz file in the specified destination.

    Args:
        filepath: the .tgz file
        folder: the destination location, if not specified then same path as source file.
    """
    if folder is None:
        folder = _os.path.dirname(filepath)
    tgz_file = _tarfile.open(filepath)
    tgz_file.extractall(path=folder)
    tgz_file.close()
    print(f"Extraction of {filepath} completed.")


def extract_zip(filepath='./data/file.zip', folder='./data') -> None:
    """
    Extracts a .zip file in the specified destination.

    Args:
        filepath: the path and filename
        folder: the destination location, if not specified then same path as source file.
    """
    if folder is None:
        folder = _os.path.dirname(filepath)
    zip_file = _zipfile.ZipFile(filepath)
    zip_file.extractall(path=folder)
    zip_file.close()


def list_dir_summary(folder='./data') -> None:
    """
    Lists a summary of the specified folder

    Args:
        folder: the folder to list the contents of
    """
    for dirpath, dirnames, filenames in _os.walk(folder):
        print(f"There are {len(dirnames)} directories, and {len(filenames)} in {dirpath}")


def load_image_as_tensor(filename: str, img_shape=(224, 224), scale=True):
    """
    Reads in an image from filename, turns it into a tensor and reshapes into
    specified shape (img_shape, img_shape, channels)

    Args:
        filename: path to target image
        img_shape: tuple with height/width dimension of target image size
        scale: scale pixel values from 0-255 to 0-1 or not

    Returns:
        Image tensor of shape (img_shape, img_shape, 3)
    """

    # Read in the image
    img = _tf.io.read_file(filename=filename)

    # Decode image into tensor
    img = _tf.io.decode_image(contents=img, channels=3)

    # # Resize the image (height / width)
    img = _tf.image.convert_image_dtype(img, _tf.float32)
    img = _tf.image.resize(images=img, size=[img_shape[0], img_shape[1]])

    if not scale:
        return _tf.cast(img * 255, _tf.int32)
    else:
        return img



import os as _os
import random as _random
import logging as _logging

import matplotlib.pyplot as _plt
import tensorflow as _tf


def load_image_dataset_from_directory(directory: str,
                                      label_mode='categorical',
                                      image_size=(256, 256),
                                      batch_size=32,
                                      class_names=None,
                                      color_mode='rgb',
                                      shuffle=True, validation_split=None):
    return _tf.keras.utils.image_dataset_from_directory(directory=directory,
                                                        label_mode=label_mode,
                                                        image_size=image_size,
                                                        batch_size=batch_size,
                                                        class_names=class_names,
                                                        color_mode=color_mode,
                                                        shuffle=shuffle,
                                                        validation_split=validation_split
                                                        )


def show_random_image_from_disk(target_dir, target_class, shape=(4, 6), cmap='gray') -> None:
    """
    Shows a random image from the file system.
    :param target_dir: The target directory, for example /food101/train
    :param target_class: The target class name, for example /steak
    :param shape: is the number of images in a grid to display
    :param cmap: is the color map to use, use "gray" for gray scale images, use None for default.
    """
    # Setup the target directory (we'll view images from here)
    try:
        target_folder = _os.path.join(target_dir, target_class)

        # Get a random image path
        fig = _plt.figure(figsize=(shape[1] * 3, shape[0] * 3))
        fig.patch.set_facecolor('gray')
        for i in range(0, shape[0] * shape[1]):
            ax = _plt.subplot(shape[0], shape[1], i + 1)
            ax.axis('off')
            image_filename = _random.sample(_os.listdir(target_folder), 1)
            image_fullpath = _os.path.join(target_folder, image_filename[0])
            img = _plt.imread(image_fullpath)
            _plt.imshow(img, cmap=cmap)
            _plt.title(f"{image_filename[0]}: {img.shape}", color='white')

        _plt.show()
    except Exception as e:
        _logging.error(e)


import itertools as _itertools

import matplotlib.pyplot as _plt
import numpy as _np
import tensorflow as _tf
import so_ml_tools as _soml
import sklearn as _sklearn


def confusion_matrix(y_true, y_prob, class_names: list[str] = None, figsize=(15, 15), text_size=10, norm=False, savefig=False) -> None:
    """
      Plots a confusion matrix of the given data.

      :param y_true: Array of truth labels (must be same shape as y_pred).
      :param y_prob: Array of probability labels (must be same shape as y_true).
      :param class_names: Array of class labels (e.g. string form). If `None`, integer labels are used.
      :param figsize: Size of output figure (default=(15, 15)).
      :param text_size: Size of output figure text (default=10).
      :param norm: normalize values or not (default=False).
      :param savefig: save confusion matrix to file (default=False).

      Plots the decision boundary created by a model predicting on X.
      Inspired by the following two websites:
      https://cs231n.github.io/neural-networks-case-study
    """

    if isinstance(y_true, _tf.data.Dataset):
        raise TypeError('y_true is a dataset, please get the labels from the dataset using '
                        '\'y_labels = ml.tf.dataset.get_labels(dataset=dataset)\'')

    y_true = _soml.util.label.to_prediction(y_prob=y_true)
    y_pred = _soml.util.label.to_prediction(y_prob=y_prob)

    # Create the confusion matrix
    cm = _sklearn.metrics.confusion_matrix(y_true, y_pred)
    cm_norm = cm.astype("float") / cm.sum(axis=1)[:, _np.newaxis]  # normalize our confusion matrix
    n_classes = cm.shape[0]

    fig, ax = _plt.subplots(figsize=figsize)

    # noinspection PyUnresolvedReferences
    cax = ax.matshow(cm, cmap=_plt.cm.Blues)
    fig.colorbar(cax)

    # Set labels to be classes
    if class_names:
        labels = class_names
    else:
        labels = _np.arange(cm.shape[0])

    # label the axes
    ax.set(title="Confusion matrix",
           xlabel="predicted label",
           ylabel="True label",
           xticks=_np.arange(n_classes),
           yticks=_np.arange(n_classes),
           xticklabels=labels,
           yticklabels=labels)

    # Set x-axis labels to bottom
    ax.xaxis.set_label_position("bottom")
    ax.xaxis.tick_bottom()

    # Adjust label size
    ax.title.set_size(text_size)

    _plt.xticks(rotation=70, fontsize=text_size)
    _plt.yticks(fontsize=text_size)

    # Set treshold for different colors
    threshold = (cm.max() + cm.min()) / 2.
    # Plot the text on each cell
    for i, j in _itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        if norm:
            _plt.text(j, i, f"{cm[i, j]} ({cm_norm[i, j] * 100:.1f}%)",
                      horizontalalignment="center",
                      color="white" if cm[i, j] > threshold else "black",
                      size=text_size)
        else:
            _plt.text(j, i, f"{cm[i, j]}",
                      horizontalalignment="center",
                      color="white" if cm[i, j] > threshold else "black",
                      size=text_size)

    if savefig:
        fig.savefig("./confusion_matrix.png")


def classification_prediction_confidence(y_true, y_pred, class_names: list[str], figsize=(10, 8)):
    _plt.figure(figsize=figsize)

    # For each column (class) in y_true
    for n in range(y_true.shape[1]):
        # Concatenate the y_true for the n'th class and y_pred for the n'th class together.
        y_combined = _np.concatenate((y_true[:, [n]], y_pred[:, [n]]), axis=1)

        # Filter out only the rows that are applicable to the n'th class (where y_true == 1)
        y_combined_class = y_combined[_np.in1d(y_combined[:, 0], [1])]

        # Delete first column (originally y_true)
        y_class = _np.delete(y_combined_class, _np.s_[0:1], axis=1)

        # Sort the values
        y_class_sorted = y_class[y_class[:, 0].argsort()] * 100

        # Plot a graph with the given values.
        class_name = str(n) if class_names is None else class_names[n]
        _plt.plot(y_class_sorted, label=class_name, linewidth=1.5)

    # X contains two features, x1 and x2
    _plt.xlabel("Predictions", fontsize=20)
    _plt.ylim([0, 100])
    _plt.yticks(_np.round(_np.arange(0, 105, 5), 1))
    _plt.ylabel(r"Confidence (%)", fontsize=20)
    _plt.legend()
    # Displaying the plot.
    _plt.show()


def classification_prediction_confidence_histogram(y_true, y_pred, class_names: list[str], figsize=(8, 4)):
    bins = range(0, 110, 10)

    fig, axs = _plt.subplots(nrows=max(int(len(class_names) / 4), 1), ncols=min(4, len(class_names)), figsize=figsize)
    axs = axs.flatten()

    # Spacing between the graphs
    fig.tight_layout(h_pad=5, w_pad=5)

    # Set the ticks for all the subplots
    _plt.setp(axs, xticks=bins)

    # For each column (class) in y_true
    for n in range(y_true.shape[1]):
        # Concatenate the y_true for the n'th class and y_pred for the n'th class together.
        y = _np.concatenate((y_true[:, [n]], y_pred[:, [n]]), axis=1)

        # Filter out only the rows that are applicable to the n'th class (where y_true == 1)
        y = y[_np.in1d(y[:, 0], [1])]

        # Delete first column (originally y_true)
        y = _np.delete(y, _np.s_[0:1], axis=1)

        # Round the values
        y = (_np.round(y, decimals=1) * 100).astype(dtype=int)

        # Plot a histogram with the given values.
        class_name = str(n) if class_names is None else class_names[n]
        axs[n].hist(y, log=True, bins=11, facecolor='#2ab0ff', edgecolor='#169acf', align='left', linewidth=0.5,
                 label=class_name)
        axs[n].set(title=class_name, xlabel='Confidence (%)', ylabel='Predictions')
    _plt.show()


def report_f1_score(y_true, y_pred, class_names: list[str], figsize=(10, 8)) -> None:
    """
    Creates a horizontal bar graph with the F1-Scores of the y_true / y_pred.
    :param y_true: Array of truth labels (must be same shape as y_pred).
    :param y_pred: Array of predicted labels (must be same shape as y_true).
    :param class_names: Array of class labels (e.g. string form). If `None`, integer labels are used.
    :param figsize: Size of output figure (default=(10, 8)).
    """
    if isinstance(y_true, _tf.data.Dataset):
        raise TypeError(
            'y_true is a dataset, please get the labels from the dataset using '
            '\'y_labels = get_labels_from_dataset(dataset=dataset, index_only=True)\'')

    y_true = _soml.util.label.to_prediction(y_prob=y_true)
    y_pred = _soml.util.label.to_prediction(y_prob=y_pred)

    # Generate classification report from SKLearn.
    report_dict = _sklearn.metrics.classification_report(y_true=y_true, y_pred=y_pred, target_names=class_names, output_dict=True)

    # Collect all the F1-scores
    f1_scores = {}
    for k, v in report_dict.items():
        if k == 'accuracy':  # Cut-off when the accuracy key is visible.
            break
        else:
            f1_scores[k] = v['f1-score']

    # Sort F1-scores from best to worst
    f1_scores_sorted = dict(sorted(f1_scores.items(), key=lambda item: item[1], reverse=True))

    _plt.rcdefaults()
    fig, ax = _plt.subplots(figsize=figsize)
    y_pos = range(len(f1_scores_sorted))
    scores = ax.barh(y_pos, f1_scores_sorted.values(), align='center')
    ax.set_yticks(y_pos)
    ax.set_yticklabels(f1_scores_sorted.keys())

    ax.invert_yaxis()  # labels read top-to-bottom
    ax.set_xlabel('F1-score')
    ax.set_title('F1-Scores')
    ax.bar_label(scores, fmt='%.2f')

    # Remove anoying white space at top / bottom.
    _plt.margins(y=0)

    _plt.show()



import tensorflow as _tf
import matplotlib.pyplot as _plt
import so_ml_tools as _soml


def get_class_names_from_dataset_info(ds_info):
    """
    Returns the labels from the dataset info object which is created from loading a
    TensorFlow Dataset (tensorflow_datasets)

    :param ds_info: The dataset info object
    :return: The labels / class names
    """
    return ds_info.features["label"].names


def get_class_names(dataset: _tf.data.Dataset):
    """
    Returns the class names from the dataset
    :param dataset: the dataset from which we want the class names.
    :return: the class names
    """
    if not isinstance(dataset, _tf.data.Dataset):
        raise TypeError('dataset is not a tf.data.Dataset')

    return dataset.class_names


def get_labels(dataset: _tf.data.Dataset):
    """
    Returns the labels from a (batched)Dataset

    :param dataset: the dataset from which we want the labels.
    :return: the labels
    """
    if not isinstance(dataset, _tf.data.Dataset):
      raise TypeError('dataset is not a tf.data.Dataset')

    input_dataset = dataset._input_dataset
    while not hasattr(input_dataset, '_batch_size') and hasattr(input_dataset, '_input_dataset'):
      input_dataset = input_dataset._input_dataset

    if hasattr(input_dataset, '_batch_size'):
      dataset = dataset.unbatch()

    y_labels = []
    for _, labels in dataset:
        y_labels.append(labels.numpy())

    return y_labels


def show_images_from_dataset(dataset: _tf.data.Dataset, shape=(4, 8)):
    assert isinstance(dataset, _tf.data.Dataset), f"The dataset supplied is not a tensorflow.data.Dataset."

    # Retrieve first batch, depending on the initalization of the dataset the batch size is default 32
    # so when performing a take of (1) we retreive the first batch
    batches = dataset.take(1)

    # Use an iterator to get the first batch of images and labels
    batch_iter = iter(batches)
    x, y_true = batch_iter.next()

    assert shape[0] * shape[1] == len(x), f"Size of shape ({shape[0]}, {shape[1]}), with a total of " \
                  f"{shape[0] * shape[1]} images, is not equal to the batch size of the dataset ({len(x)})."

    if len(x) != shape[0] * shape[1]:
        raise TypeError('dataset is not a Dataset')

    _soml.data.image.show_images_from_nparray_or_tensor(x=x, y=y_true, class_names=dataset.class_names, shape=shape)
    _plt.show()



import pandas as _pd
import sklearn as _sklearn
import tensorflow as _tf
import so_ml_tools as _soml


def determine_outliers(x, y_true, y_pred = None, y_prob = None, top=10) -> _pd.DataFrame:
    """
    Displays the top X (default 10) outliers between y_true and y_pred.

    Args:
        x: data to display
        y_true: actual labels [3, 3, 2, ..., 4, 4, 1]
        y_pred: predictions [3, 3, 2, ..., 4, 2, 1]
        y_prob: probabilities (multiclass) [[0.30974227, ...], [0.32494593, ...], ..]
        top: how many outliers to return

    Returns:
        The outliers
    """

    # Check if we need to convert y_prob to numpy.
    _y_prob = y_prob
    if _tf.is_tensor(x=y_prob):
        _y_prob = y_prob.numpy()

    # If y_pred is None then determine the predictions, otherwise check if we need to convert to numpy.
    _y_pred = y_pred
    if y_pred is None:
        _y_pred = _soml.util.label.to_prediction(y_prob=_y_prob)
    elif _tf.is_tensor(x=y_pred):
        _y_pred = y_pred.numpy()

    # Create a matrix containing the y_true, y_pred and y_prob value in columns.
    data = None
    if _soml.util.label.is_multiclass_classification(_y_prob):
        data = [[y_true[i], x, _y_prob[i][x]] for i, x in enumerate(_y_pred)]
    elif _soml.util.label.is_binary_classification(_y_prob):
        data = [[y_true[i], x[0], _y_prob[i][0]] for i, x in enumerate(_y_pred)]

    if data is None:
        raise ValueError('y_prob is not a multiclass or binary classification.')

    # Create a dataframe from the matrix with the appropriate column names
    data_df = _pd.DataFrame(data, columns=['y_true', 'y_pred', 'y_prob'])

    # Add the columns of X to the dataframe.
    all_df = data_df
    if x is not None:
        all_df = _pd.concat([x, data_df], axis=1)

    # From all rows that have in-equal y_true vs y_pred, sort on y_prob in descending and select the :top rows.
    outliers = all_df[all_df['y_true'] != all_df['y_pred']].sort_values('y_prob', ascending=False)[:top]

    # Round the y_prob column to 2 decimals, for that we need to cast the y_prob column to float instead of float16
    outliers['y_prob'] = outliers['y_prob'].astype(float).round(decimals=2)

    return outliers


def quality_metrics(y_true, y_pred):
    """
    calculates model accuracy, precision, recall and F1-Score
    :param y_true: the truth labels
    :param y_pred: the predictions
    :return: dictionary containing accuracy, precision, recall, f1 score and support
    """
    # Calculate model accuracy
    model_accuracy = _sklearn.metrics.accuracy_score(y_true, y_pred) * 100

    # Calculate precision, recall and F1-score using "weighted" average,
    # weighted will also take the amount of samples for each in mind.
    model_precission, model_recall, model_f1_score, support = \
        _sklearn.metrics.precision_recall_fscore_support(y_true, y_pred, average="weighted")
    model_results = \
        {
            "accuracy" : model_accuracy,
            "precision": model_precission,
            "recall": model_recall,
            "f1-score": model_f1_score,
            "support": support
        }
    return dict(sorted(model_results.items()))


import pandas as _pd
import so_ml_tools as _soml


def determine_outliers(x, y_true, y_pred, target_column: str, top=10):
    """
    Displays the top X (default 10) outliers between y_true and y_pred.
    :param x: data to display
    :param y_true: labels
    :param y_pred: predictions
    :param target_column: column on which we can calculate the mean negative / positive.
    :param top: how many rows to display
    :return: dataframe containing the top X.
    """
    x_copy = x.copy()
    x_copy["y_true"] = y_true.round(0)
    x_copy["y_pred"] = y_pred.round(0)

    diff = y_true - y_pred
    x_copy["diff"] = diff.round(0)

    mean_positive = diff[diff[target_column] > 0].mean()[0]
    mean_negative = diff[diff[target_column] < 0].mean()[0]

    positive_outliers = x_copy[x_copy["diff"] > mean_positive]
    negative_outliers = x_copy[x_copy["diff"] < mean_negative]

    positive_outliers_sorted = positive_outliers.sort_values(['diff'], ascending=False, inplace=False)[:top]
    negative_outliers_sorted = negative_outliers.sort_values(['diff'], ascending=True, inplace=False)[:top]

    return _pd.concat([positive_outliers_sorted, negative_outliers_sorted])


def quality_metrics(y_true, y_pred):
    """
    calculates model accuracy, precision, recall and F1-Score
    :param y_true: the truth labels
    :param y_pred: the predictions
    :return: dictionary containing accuracy, precision, recall and f1 score.
    """
    return _soml.multiclass.score.quality_metrics(y_true, y_pred)


import matplotlib.pyplot as _plt
import numpy as _np
import tensorflow as _tf

import so_ml_tools as _soml

from scipy.interpolate import interp1d as _interp1d


def confusion_matrix(y_true, y_prob, class_names: list[str] = None, figsize=(15, 15), text_size=10, norm=False, savefig=False) -> None:
    """
      Plots a confusion matrix of the given data.

      :param y_true: Array of truth labels (must be same shape as y_pred).
      :param y_prob: Array of probability labels (must be same shape as y_true).
      :param class_names: Array of class labels (e.g. string form). If `None`, integer labels are used.
      :param figsize: Size of output figure (default=(15, 15)).
      :param text_size: Size of output figure text (default=10).
      :param norm: normalize values or not (default=False).
      :param savefig: save confusion matrix to file (default=False).

      Plots the decision boundary created by a model predicting on X.
      Inspired by the following two websites:
      https://cs231n.github.io/neural-networks-case-study
    """
    _soml.multiclass.plot.confusion_matrix(y_true, y_prob, class_names, figsize, text_size, norm, savefig)


def classification_prediction_confidence(y_true, y_pred, class_names: list[str], figsize=(10, 8)):
    _plt.figure(figsize=figsize)

    # In binary classification we have two possible best outcomes, 0 and 1, both ends mean that something has
    # been positively identified, so we need to first extract the two classes based on y_true (0 is the first class
    # and 1 is the second class), then we need to translate the value ranges, for results
    # of class 1 (y_true[:] == 1) the closer the number to 1 the higher de confidence, and for results
    # of class 0 (y_true[:] == 0) the close the number to 0 the higher de confidence.
    # For this we use the interp1d function.
    positivate_range_translation = _interp1d([0.,1.],[0,100])
    negative_range_translation = _interp1d([0.,1.],[100,0])

    y_true_int = _soml.util.label.probability_to_binary(y_true, dtype=_np.int32)
    for n in range(2):
        # Filter out only the rows thar are applicatie to the n'th class
        y_pred_single_class = y_pred[_np.in1d(y_true_int[:, 0], [n])]

        # Convert ranges as described above
        if n == 1:
            y_pred_single_class_translated = positivate_range_translation(y_pred_single_class)
        else:
            y_pred_single_class_translated = negative_range_translation(y_pred_single_class)

        # Sort the values
        y_pred_single_class_translated_sorted = y_pred_single_class_translated[y_pred_single_class_translated[:, 0].argsort()]

        # Plot a graph with the given values.
        class_name = str(n) if class_names is None else class_names[n]
        _plt.plot(y_pred_single_class_translated_sorted, label=class_name, linewidth=1.5)

    # X contains two features, x1 and x2
    _plt.xlabel("Predictions", fontsize=20)
    _plt.ylim([0, 100])
    _plt.yticks(_np.round(_np.arange(0, 105, 5), 1))
    _plt.ylabel(r"Confidence (%)", fontsize=20)
    _plt.legend()
    # Displaying the plot.
    _plt.show()


def classification_prediction_confidence_histogram(y_true, y_pred, class_names: list[str], figsize=(8, 4)):
    bins = range(0, 110, 10)

    fig, axs = _plt.subplots(nrows=1, ncols=2, figsize=figsize)

    # Spacing between the graphs
    fig.tight_layout(h_pad=5, w_pad=5)

    # Set the ticks for all the subplots
    _plt.setp(axs, xticks=bins)

    # In binary classification we have two possible best outcomes, 0 and 1, both ends mean that something has
    # been positively identified, so we need to first extract the two classes based on y_true (0 is the first class
    # and 1 is the second class), then we need to translate the value ranges, for results
    # of class 1 (y_true[:] == 1) the closer the number to 1 the higher de confidence, and for results
    # of class 0 (y_true[:] == 0) the close the number to 0 the higher de confidence.
    # For this we use the interp1d function.
    positivate_range_translation = _interp1d([0.,1.],[0,100])
    negative_range_translation = _interp1d([0.,1.],[100,0])

    y_true_int = _soml.util.label.probability_to_binary(y_true, dtype=_np.int32)
    for n in range(2):
        # Filter out only the rows thar are applicatie to the n'th class
        y_pred_single_class = y_pred[_np.in1d(y_true_int[:, 0], [n])]

        # Convert ranges as described above
        if n == 1:
            y_pred_single_class_translated = positivate_range_translation(y_pred_single_class)
        else:
            y_pred_single_class_translated = negative_range_translation(y_pred_single_class)

        # Sort the values
        y = (_np.round(y_pred_single_class_translated / 10) * 10).astype(dtype=int)

        # Plot a graph with the given values.
        class_name = str(n) if class_names is None else class_names[n]
        axs[n].hist(y, log=True, bins=11, facecolor='#2ab0ff', edgecolor='#169acf', align='left', linewidth=0.5,
                    label=class_name)
        axs[n].set(title=class_name, xlabel='Confidence (%)', ylabel='Predictions')
    _plt.show()


def report_f1_score(y_true, y_pred, class_names: list[str], figsize=(10, 8)) -> None:
    _soml.multiclass.plot.report_f1_score(y_true, y_pred, class_names, figsize)


def xy_data_with_label(x, y) -> None:
    """
    Plots a graph of the values of x whereas x contains vectors of x/y coordinates and y
    is the label (0 or 1).

    :param x: is an array containing vectors of x/y coordinates.
    :param y: are the associated labels (0=blue, 1=red)
    """
    _plt.plot(x[:, 0][y == 1], x[:, 1][y == 1], "bs")
    _plt.plot(x[:, 0][y == 0], x[:, 1][y == 0], "r^")

    # X contains two features, x1 and x2
    _plt.xlabel(r"$x_1$", fontsize=20)
    _plt.ylabel(r"$x_2$", fontsize=20)

    # Displaying the plot.
    _plt.show()


def decision_boundary(model: _tf.keras.Model, x, y) -> None:
    """
    Plots the decision boundary created by a model predicting on X.

    Inspired by the following two websites:
    https://cs231n.github.io/neural-networks-case-study

    :param model: the sequence model.
    :param x: array containing vectors with x/y coordinates
    :param y: are the associated labels (0=blue, 1=red)
    """
    # Define the axis boundaries of the plot and create a meshgrid.
    x_min, x_max = x[:, 0].min() - 0.1, x[:, 0].max() + 0.1
    y_min, y_max = x[:, 1].min() - 0.1, x[:, 1].max() + 0.1

    xx, yy = _np.meshgrid(_np.linspace(x_min, x_max, 100),
                         _np.linspace(y_min, y_max, 100))

    # Create X value (we're going to make predictions on these)
    x_in = _np.c_[xx.ravel(), yy.ravel()]  # Stack 2D arrays together

    # Make predictions
    y_pred = model.predict(x_in)

    # Check for multi-class
    if len(y_pred[0]) > 1:
        print("doing multiclass classification")
        # We have to reshape our predictions to get them ready for plotting.
        y_pred = _np.argmax(y_pred, axis=1).reshape(xx.shape)
    else:
        print("doing binary classification")
        y_pred = _np.round(y_pred).reshape(xx.shape)

    # Plot the decision boundary
    # noinspection PyUnresolvedReferences
    _plt.contourf(xx, yy, y_pred, cmap=_plt.cm.RdYlBu, alpha=0.7)
    # noinspection PyUnresolvedReferences
    _plt.scatter(x[:, 0], x[:, 1], c=y, s=40, cmap=_plt.cm.RdYlBu)
    _plt.xlim(xx.min(), xx.max())
    _plt.ylim(yy.min(), yy.max())


def y_pred_vs_y_true(y_true, y_pred, figsize=(10, 8)) -> None:
    """
    Shows a graph with the predicted values vs the truth labels.
    :param y_true: the truth labels
    :param y_pred: the predicted values
    :param figsize: size of the graph
    """
    # Merge the two columns into a single new numpy array.
    m = _np.append(y_true.round(2), y_pred.round(2), axis=1)

    # Add a third column with the difference
    m_a = _np.append(m, y_true - y_pred, axis=1)

    # sort based on y_true
    s = m_a[m_a[:,0].argsort()]
    _plt.figure(figsize=figsize, facecolor='#FFFFFF')
    _plt.plot(s[:,0], label="y_true", color="#0000FF", linestyle="solid", linewidth=1.5)
    _plt.plot(s[:,1], label="y_pred", color="#FF0000", linestyle="solid", linewidth=1.5)
    _plt.plot(s[:,2], label="diff", color="#FF0000", linestyle="solid", linewidth=1.5)
    _plt.title('y_true vs y_pred with difference', size=20)
    _plt.xlabel('Predictions', size=14)
    _plt.ylabel('Value', size=14)
    _plt.legend()


from . import (
    data,
    multiclass,
    nlp,
    pandas,
    regression,
    tf,
    util
)

# We follow Semantic Versioning (https://semver.org/spec/v2.0.0.html)
_MAJOR_VERSION = '0'
_MINOR_VERSION = '1'
_PATCH_VERSION = '10'

__version__ = '.'.join([
    _MAJOR_VERSION,
    _MINOR_VERSION,
    _PATCH_VERSION,
])

__all__ = [
    "data",
    "multiclass",
    "nlp",
    "pandas",
    "regression",
    "tf",
    "util",
    "__version__"
]


